﻿Imports System.Drawing.Drawing2D

Module GraphPaperModule
    Friend TimeGraphRefreshFlag As Boolean = True
    Friend GraphPaper1 As GraphPaper = New GraphPaper
End Module

Public Class GraphPaper

    Friend PB_Graph As PictureBox

    Friend GraphType As Int32

    Private BackColor As Color = Color.FromArgb(255, 250, 240)                  '255, 250, 240
    Private BackColor2 As Brush = New SolidBrush(Color.FromArgb(255, 252, 250)) '255, 252, 250  

    Private Pen_Scale1 As Pen = New Pen(Color.FromArgb(150, 120, 120))
    Private Pen_Scale2 As Pen = New Pen(Color.FromArgb(200, 200, 200))
    Private Pen_Scale3 As Pen = New Pen(Color.FromArgb(220, 220, 220))

    Private Pen_Trace1 As Pen = New Pen(Color.FromArgb(60, 60, 180), 4)
    Private Pen_Trace2 As Pen = New Pen(Color.FromArgb(0, 150, 0), 4)
    Private Pen_Trace3 As Pen = New Pen(Color.FromArgb(150, 0, 0), 4)

    Private Brush_Trace1 As Brush = New SolidBrush(Color.FromArgb(40, 40, 180))
    Private Brush_Trace2 As Brush = New SolidBrush(Color.FromArgb(0, 120, 0))
    Private Brush_Trace3 As Brush = New SolidBrush(Color.FromArgb(150, 40, 40))

    Private Brush_Text As Brush = New SolidBrush(Color.FromArgb(50, 50, 50))
    Private Brush_HilightBack As Brush = New SolidBrush(Color.FromArgb(150, 255, 255, 120))
    Private Brush_HilightText As Brush = New SolidBrush(Color.FromArgb(60, 60, 60))

    Private Font_Text As Font = New Font("Arial", 10, FontStyle.Bold)
    Private Font_Scale As Font = New Font("Arial", 9, FontStyle.Bold)
    Private Font_Labels As Font = New Font("Arial", 10, FontStyle.Bold)
    Private Font_Messages As Font = New Font("Arial", 14, FontStyle.Bold)

    Private hpix As Int32
    Private wpix As Int32

    Private BorderLeft As Int32 = 44
    Private BorderRight As Int32 = 36
    Private BorderUp As Int32 = 32
    Private BorderDown As Int32 = 32

    Private xmin As Double
    Private xmax As Double
    Private xstep As Double

    Private ymin As Double
    Private ymax As Double
    Private ystep As Double

    ' =====================================================================================================
    '  SCALE CONVERSION FUNCTIONS
    ' =====================================================================================================
    Private Function ValueToX(ByVal value As Double) As Single
        Dim wgraph As Int32 = wpix - BorderLeft - BorderRight
        value = (value - xmin) / (xmax - xmin)
        If Double.IsNaN(value) Then value = 0
        Return CSng(BorderLeft + value * wgraph)
    End Function
    Private Function XToValue(ByVal X As Double) As Double
        Dim wgraph As Int32 = wpix - BorderLeft - BorderRight
        Return (Xmax * (X - BorderLeft) + Xmin * (wpix - BorderRight - X)) / wgraph
    End Function

    Private Function ValueToY(ByVal value As Double) As Single
        Dim hgraph As Int32 = hpix - BorderUp - BorderDown
        value = (value - ymin) / (ymax - ymin)
        If Double.IsNaN(value) Then value = 0
        Return CSng(hpix - BorderDown - (value * hgraph))
    End Function
    Private Function YToValue(ByVal Y As Double) As Double
        Dim hgraph As Int32 = hpix - BorderUp - BorderDown
        Return (Ymax * (hpix - BorderDown - Y) + Ymin * (Y - BorderUp)) / hgraph
    End Function

    ' ==============================================================================
    '  INIT X and Y MIN/MAX/STEP
    ' ==============================================================================
    Private Sub InitXminMaxStep()
        xmin = Axes(0).Min
        xmax = Axes(0).Max
        xstep = Math.Abs(Axes(0).Stp)
        '
        If Double.IsNaN(xmin) Then
            xmin = Axes(0).ValueMin
        End If
        If Double.IsNaN(xmax) Then
            xmax = Axes(0).ValueMax
        End If
        '
        If xmax = xmin Then
            xmin = Axes(0).ValueMin
            xmax = Axes(0).ValueMax
        End If
        '
        If xstep = 0 Or Double.IsNaN(xstep) Or xstep * 30 < Math.Abs(xmax - xmin) Then
            xstep = Math.Abs(xmax - xmin) * 100 / (wpix - BorderLeft - BorderRight)
            xstep = XaxisTimeUnit * RoundValueTo_1_2_5(xstep / XaxisTimeUnit)
        End If
        '
        If xstep = 0 Or xmin = xmax Then ' Or xstep * 30 < Math.Abs(xmax - xmin) 
            xmin = 0
            xmax = 100
            xstep = 10
        End If
        '
        xstep = xstep * Math.Sign(xmax - xmin)
        '
        'Debug.Print(xstep.ToString)
    End Sub

    Private Sub InitYminMaxStep(ByVal a As Int32)
        ymin = Axes(a).Min
        ymax = Axes(a).Max
        ystep = Math.Abs(Axes(a).Stp)
        '
        If Double.IsNaN(ymin) Then
            ymin = Axes(a).ValueMin
        End If
        If Double.IsNaN(ymax) Then
            ymax = Axes(a).ValueMax
        End If
        '
        If ymax = ymin Then
            ymin = Axes(a).ValueMin
            ymax = Axes(a).ValueMax
        End If
        '
        If ystep = 0 Or Double.IsNaN(ystep) Or ystep * 30 < Math.Abs(ymax - ymin) Then
            ystep = Math.Abs(ymax - ymin) * 45 / (hpix - BorderDown)
            ystep = RoundValueTo_1_2_5(ystep)
        End If
        '
        If ystep = 0 Or ymin = ymax Then  'Or ystep * 30 < Math.Abs(ymax - ymin)
            ymin = 0
            ymax = 100
            ystep = 10
        End If
        '
        ystep = ystep * Math.Sign(ymax - ymin)
    End Sub

    ' ==============================================================================
    '  SCALES
    ' ==============================================================================
    Private Sub DrawScaleX(ByVal g As Graphics)
        ' --------------------------------------------------------------------------- 
        InitXminMaxStep()
        ' ---------------------------------------------------------------------------
        Dim px As Single

        ' TODO ----- XaxisIsTime - DrawScaleX
        If XaxisIsTime Then
            For x As Double = xmax To xmin - 0.000001 Step -xstep
                px = ValueToX(x)
                g.DrawLine(Pen_Scale2, px, BorderUp, px, hpix - BorderDown)
                px = Math.Max(px, 10)
                g.DrawString(CSng((xmax - x) / XaxisTimeUnit).ToString("0.###", GCI), Font_Scale, Axes(0)._Brush, px - 9, hpix - BorderDown + 4)
            Next
        Else
            For x As Double = xmin To xmax + 0.000001 Step xstep
                px = ValueToX(x)
                g.DrawLine(Pen_Scale2, px, BorderUp, px, hpix - BorderDown)
                px = Math.Max(px, 10)
                g.DrawString(CSng(x / XaxisTimeUnit).ToString("0.###", GCI), Font_Scale, Axes(0)._Brush, px - 9, hpix - BorderDown + 4)
            Next
        End If
        Dim p As Single = g.MeasureString(Axes(0).Name, Font_Scale).Width
        g.DrawString(Axes(0).Name, Font_Scale, Axes(0)._Brush, wpix - BorderRight - p - 4, hpix - BorderDown + 16)
        g.DrawLine(Pen_Scale1, wpix - BorderRight, BorderUp, wpix - BorderRight, hpix - BorderDown)
        g.DrawLine(Pen_Scale1, BorderLeft, BorderUp, BorderLeft, hpix - BorderDown)
    End Sub

    Private Sub DrawScaleY(ByVal g As Graphics)
        For a As Int32 = Axes.Length - 1 To 1 Step -1
            ' --------------------------------------------------------------------------- 
            InitYminMaxStep(a)
            ' ---------------------------------------------------------------------------
            Dim b As SolidBrush = Axes(a)._Brush
            ' --------------------------------------------------------------------------- Position X
            Dim px As Single = 4
            If a > LeftColumnsCount Then
                px = wpix - BorderRight + ColumnSize * (a - LeftColumnsCount - 1) + 4
            Else
                px = ColumnSize * (a - 1) + 4
            End If
            Axes(a).PosX = px
            ' -------------------------------------------------------------------------- ymin round to ndigits
            ymin = RoundToNdigits(ymin, NumDigits)
            ' -------------------------------------------------------------------------- scale values
            Dim py As Single
            For y As Double = ymin To ymax + 0.000001 Step ystep
                py = ValueToY(y)
                If a = 1 Then g.DrawLine(Pen_Scale2, BorderLeft, py, wpix - BorderRight, py)
                g.DrawString(RoundToNdigits(y, NumDigits).ToString(GCI), Font_Scale, b, px, py - 8)
            Next
            g.DrawString(Axes(a).Name, Font_Scale, b, px, 4)
            ' --------------------------------------------------------------------------- selected track rectangle
            px -= 2
            If SelectedTrack = a Then
                g.DrawRectangle(Pens.DarkRed, px, 2, ColumnSize, hpix - BorderDown + 8)
            Else
                g.DrawLine(Pen_Scale3, px, 2, px, hpix - BorderDown + 6)
            End If
        Next
        g.DrawLine(Pen_Scale1, BorderLeft, BorderUp - 1, wpix - BorderRight, BorderUp - 1)
        g.DrawLine(Pen_Scale1, BorderLeft, hpix - BorderDown + 1, wpix - BorderRight + 1, hpix - BorderDown + 1)
    End Sub

    ' ==============================================================================
    '  GRAPH
    ' ==============================================================================
    Private Sub UpdateTrackBars()
        ' --------------------------------------------------------------------- TrackBar position and size
        Form1.TrackBar1.Left = BorderLeft + 11
        Form1.TrackBar1.Width = wpix - BorderLeft - BorderRight - 1
        Form1.TrackBar1.Visible = XaxisIsTime
        '
        Form1.TrackBar2.Left = 14
        Form1.TrackBar2.Width = BorderLeft - 10
        Form1.TrackBar2.Visible = True
        '
        Form1.TrackBar3.Left = wpix - BorderRight + 16
        Form1.TrackBar3.Width = BorderRight - 12
        Form1.TrackBar3.Visible = True
    End Sub

    Private Sub Message(ByVal g As Graphics, ByVal s As String)
        g.DrawString(s, Font_Messages, Brushes.Black, 16, 16)
    End Sub

    Private Sub Do_Axes_OldValue_Smoothing()
        ' ----------------------------------------------------------- ReDim DataRowsSmoothed array
        ReDim DataRowsSmoothed(DataRows.GetUpperBound(0), _
                               DataRows.GetUpperBound(1))
        ' ----------------------------------------------------------- Smoothing step variable
        Dim Stp As Int32 = 1
        Dim SmoothingStp As Double = 1
        If Form1.FastDrawing Then
            Stp = NumDataLines \ 1000
            If Stp < 1 Then Stp = 1
            SmoothingStp = Math.Sqrt(Stp)
        End If
        ' -----------------------------------------------------------
        Dim y_column As Int32
        For a As Int32 = Axes.Length - 1 To 1 Step -1
            With Axes(a)
                ' ------------------------------------------------------------- Init
                y_column = Math.Min(.Column, NumDataColumns - 1)
                .OldValue = DataRows(0, y_column)
                .ValueMin = Double.MaxValue
                .ValueMax = Double.MinValue
                ' ------------------------------------------------------------- Limit smoothing
                If .Smoothing < SmoothingStp Then .Smoothing = SmoothingStp
                ' ------------------------------------------------------------- Smoothing start point
                Dim SmoothingStartPoint As Int32
                SmoothingStartPoint = Visible_NumLines \ 500
                ' ------------------------------------------------------------- FOR each point
                For i As Int32 = 0 To NumDataLines - 1 Step Stp
                    ' --------------------------------------------------------- Smoothing
                    Dim v As Double = DataRows(i, y_column)
                    If i > SmoothingStartPoint And .Smoothing >= 1 Then
                        .OldValue += (v - .OldValue) / .Smoothing * SmoothingStp
                        DataRowsSmoothed(i, y_column) = .OldValue
                        ' ----------------------------------------------------- Update ValueMin and ValueMax
                        .ValueMin = Math.Min(.ValueMin, .OldValue)
                        .ValueMax = Math.Max(.ValueMax, .OldValue)
                    Else
                        DataRowsSmoothed(i, y_column) = v
                        .OldValue = v
                    End If
                Next
                ' -------------------------------------------------------------
                If Double.IsNaN(.ValueMin) Then .ValueMin = 0
                If Double.IsNaN(.ValueMax) Then .ValueMax = 100
                If .ValueMin = Double.MaxValue Then .ValueMin = 0
                If .ValueMax = Double.MinValue Then .ValueMax = 100
                ' -------------------------------------------------------------
                InitYminMaxStep(a)
            End With
        Next
    End Sub

    Friend Sub UpdateGraph()
        ' --------------------------------------------------------------------- 90 mS for longTest2 431K - 5400 lines
        'Dim sw1 As Stopwatch = New Stopwatch : sw1.Start()
        ' ---------------------------------------------------------------------
        'On Error Resume Next
        ' --------------------------------------------------------------------- 
        If Not PB_Graph.Visible Then Return
        InitPictureboxImage(PB_Graph)
        If PB_Graph.Image Is Nothing Then Return
        ' ---------------------------------------------------------------------
        hpix = PB_Graph.Image.Height
        wpix = PB_Graph.Image.Width
        If hpix < 1 Or wpix < 1 Then Return
        ' ---------------------------------------------------------------------
        Dim g As Graphics = Graphics.FromImage(PB_Graph.Image)
        g.Clear(BackColor)
        ' ---------------------------------------------------------------------
        If Axes.Length = 0 Then Message(g, "Invalid axis length") : Return
        If NumDataLines < 1 Then Message(g, "Invalid data file." + vbCrLf + vbCrLf + "Valid lines are less then one.") : Return
        If NumDataColumns < 2 Then Message(g, "Invalid data file." + vbCrLf + vbCrLf + "Waiting for some valid line.") : Return
        ' --------------------------------------------------------------------- Prepare BorderLeft and BorderRight
        If LeftColumnsCount < 0 Then LeftColumnsCount = 0
        If LeftColumnsCount > Axes.Length - 1 Then LeftColumnsCount = Axes.Length - 1
        BorderLeft = (LeftColumnsCount * ColumnSize) + 6
        BorderRight = ColumnSize * (Axes.Length - LeftColumnsCount - 1) + 10
        If wpix - BorderLeft - BorderRight < 0 Then BorderRight = wpix - BorderLeft
        ' --------------------------------------------------------------------- Do the SMOOTHING
        Do_Axes_OldValue_Smoothing()
        ' --------------------------------------------------------------------- TrackBars position and size
        UpdateTrackBars()
        ' --------------------------------------------------------------------- draw scales
        g.FillRectangle(BackColor2, BorderLeft, BorderUp, _
                        wpix - BorderLeft - BorderRight, _
                        hpix - BorderUp - BorderDown)
        DrawScaleX(g)
        DrawScaleY(g)
        ' --------------------------------------------------------------------- 
        g.SmoothingMode = Drawing2D.SmoothingMode.HighQuality
        ' ---------------------------------------------------------------------
        Dim oldx As Single
        Dim oldy As Single
        Dim x As Single
        Dim y As Single
        Dim x_column As Int32
        Dim y_column As Int32
        Dim v As Double
        ' --------------------------------------------------------------------- Step variable
        Dim Stp As Int32 = 1
        If Form1.FastDrawing Then
            Stp = NumDataLines \ 1000
            If Stp < 1 Then Stp = 1
        End If
        ' ---------------------------------------------------------------------
        x_column = Math.Min(Axes(0).Column, NumDataColumns - 1)
        ' ---------------------------------------------------------------------
        For a As Int32 = Axes.Length - 1 To 1 Step -1
            y_column = Math.Min(Axes(a).Column, NumDataColumns - 1)
            InitYminMaxStep(a)
            oldx = 0
            ' ----------------------------------------------------------------- FOR each point
            For i As Int32 = 0 To NumDataLines - 1 Step Stp
                x = ValueToX(DataRows(i, x_column)) - 2
                v = DataRowsSmoothed(i, y_column)
                y = ValueToY(v)
                ' -------------------------------------------------------------
                If x >= BorderLeft And x <= wpix - BorderRight And y >= BorderUp And y <= hpix - BorderDown Then
                    If SelectedTrack >= 0 Then
                        If SelectedTrack = a Then
                            g.FillRectangle(Axes(a)._Brush, x - 3, y - 3, 5, 5)
                        End If
                        If oldx > BorderLeft AndAlso x > BorderLeft Then
                            g.DrawLine(Axes(a)._Pen, oldx, oldy, x, y)
                        End If
                    Else
                        Select Case GraphType
                            Case 0
                                If oldx > BorderLeft AndAlso x > BorderLeft Then
                                    g.DrawLine(Axes(a)._Pen, oldx, oldy, x, y)
                                End If
                            Case 1
                                g.FillRectangle(Axes(a)._Brush, x - 1, y - 1, 3, 3)
                            Case 2
                                g.FillRectangle(Axes(a)._Brush, x - 3, y - 3, 5, 5)
                        End Select
                    End If
                    oldx = x
                    oldy = y
                End If
            Next
        Next
        ' --------------------------------------------------------------------- show
        PB_Graph.Invalidate()
        'ReportString = "Scope time: " + (sw1.Elapsed.TotalMilliseconds * 1000).ToString("0") & " uS"
        'Form1.Text = "Scope time: " + (sw1.Elapsed.TotalMilliseconds * 1000).ToString("0") & " uS"
        'Form1.Text = "Scope time: " + (sw1.Elapsed.TotalMilliseconds).ToString("0")
    End Sub

    'Private Sub DrawHilightedString(ByVal g As Graphics, _
    '                                ByVal str As String, _
    '                                ByVal br As Brush, _
    '                                ByVal x As Single, _
    '                                ByVal y As Single, _
    '                                Optional ByVal align As HorizontalAlignment = HorizontalAlignment.Left)
    '    Dim sz As SizeF = g.MeasureString(str, Font_Labels, 999)
    '    Select Case align
    '        Case HorizontalAlignment.Left
    '        Case HorizontalAlignment.Center : x -= sz.Width / 2.0F
    '        Case HorizontalAlignment.Right : x -= sz.Width
    '    End Select
    '    g.FillRectangle(Brush_HilightBack, x - 6, y - 2, sz.Width + 6, sz.Height + 4)
    '    g.DrawString(str, Font_Labels, br, x, y)
    'End Sub


    ' ==============================================================================
    '  SHOW TRACK INFO
    ' ==============================================================================
    Friend TrackInfo As String = ""
    Friend SelectedTrack As Int32 = -1

    Friend Sub PrepareTrackInfo(ByVal mouseX As Int32, ByVal mouseY As Int32)
        '
        Dim x As Single
        Dim y As Single
        Dim x_column As Int32
        Dim y_column As Int32
        Dim Dist As Double
        Dim MinDist As Double = 99999
        Dim v As Double
        '
        If NumDataLines = 0 Or NumDataColumns = 0 Then Return
        '
        If mouseY < BorderUp Then
            TrackInfo = ""
        ElseIf mouseY < hpix - BorderDown Then
            If mouseX < BorderLeft Or mouseX > wpix - BorderRight Then
                SelectedTrack = GetAxisIndexFromXposition(mouseX)
                TrackInfo = ""
            Else
                SelectedTrack = -1
                ' ------------------------------------------------------------ step variable
                Dim Stp As Int32 = 1
                If Form1.FastDrawing Then
                    Stp = NumDataLines \ 1000
                    If Stp < 1 Then Stp = 1
                End If
                ' -------------------------------------------------------------
                x_column = Math.Min(Axes(0).Column, NumDataColumns - 1)
                ' -------------------------------------------------------------
                For a As Int32 = Axes.Length - 1 To 1 Step -1
                    y_column = Math.Min(Axes(a).Column, NumDataColumns - 1)
                    InitYminMaxStep(a)
                    ' --------------------------------------------------------- FOR each point
                    For i As Int32 = 0 To NumDataLines - 1 Step Stp
                        x = ValueToX(DataRows(i, x_column))
                        v = DataRowsSmoothed(i, y_column)
                        y = ValueToY(v)
                        ' ------------------------------------------------------ Mindist
                        Dist = Math.Abs(x - mouseX) + Math.Abs(y - mouseY)
                        If Dist < MinDist Then
                            MinDist = Dist
                            TrackInfo = Axes(a).Name + " " + v.ToString("0.###", GCI)
                            SelectedTrack = a
                        End If
                    Next
                Next
                If MinDist > 30 Then
                    TrackInfo = ""
                    SelectedTrack = -1
                End If
            End If
        Else
            ' ----------------------------------------------------------------- AXIS X - INFO
            If mouseX > BorderLeft And mouseX < wpix - BorderRight Then
                Dim kx As Double
                Dim line As Int32
                kx = (mouseX - BorderLeft) / (wpix - BorderLeft - BorderRight)
                line = FirstValidLine + CInt((DataLines.Length - FirstValidLine) * kx)
                If line < FirstValidLine Then line = FirstValidLine
                If line >= DataLines.Length - 1 Then line = DataLines.Length - 1
                TrackInfo = Left(DataLines(line), 19)
            Else
                TrackInfo = ""
            End If
            SelectedTrack = -1
        End If
    End Sub

    Private Function GetAxisIndexFromXposition(ByVal x As Single) As Int32
        For a As Int32 = Axes.Length - 1 To 1 Step -1
            If x > Axes(a).PosX Then
                Return a
            End If
        Next
        Return -1
    End Function

End Class

