﻿
Public Class SimpleRTB
    Inherits System.Windows.Forms.RichTextBox

    ' ===============================================================================================
    '  Reformat Text
    ' ===============================================================================================
    Friend RTB_Lines() As String

    Private TabsArray() As Integer = {100, 200, 300, 400, 500, 600, 700}

    Friend Sub InitWithText(ByVal txt As String)
        Me.SelectionTabs = TabsArray
        EventsAreEnabled = False
        Me.Text = txt
        ColorLines()
        EventsAreEnabled = True
        UndoStack.Clear()
        RedoStack.Clear()
    End Sub

    Private Sub ReformatText()
        ' ---------------------------------------------------------- init load time
        'Dim sw As Stopwatch = New Stopwatch
        'sw.Start()
        ' ---------------------------------------------------------- init
        RTB_Lines = Me.Lines
        ' ---------------------------------------------------------- 
        Dim oldPaintEvent As Boolean = PaintEventEnabled
        Dim oldEnabled As Boolean = Me.Enabled
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------- save position 
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        ' ----------------------------------------------------------
        If RTB_Lines.Length < 2 Then ReDim RTB_Lines(1)
        RTB_Lines(0) = "Axis	Name	Col.	Min	Max	Step	Color     "
        RTB_Lines(1) = StrDup(70, "-")
        Me.Lines = RTB_Lines
        ' ---------------------------------------------------------- set positions to tabs
        For i As Int32 = 2 To RTB_Lines.Length - 1
            Dim s As String = ReplaceMultipleSpacesAndTabAndTrim(RTB_Lines(i))
            s = s.Replace(" ", vbTab)
            If s.Split(CChar(vbTab)).Length = 7 Then
                RTB_Lines(i) = s
            End If
        Next
        Me.Lines = RTB_Lines
        ' ---------------------------------------------------------- restore
        RestoreSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        PaintEventEnabled = oldPaintEvent
        Me.Enabled = oldEnabled
        Me.Focus()
        ' ---------------------------------------------------------- load time
        'ReformatTextMilliseconds = sw.ElapsedMilliseconds
    End Sub

    Private Sub ColorLines()
        ' ---------------------------------------------------------- 
        Dim oldPaintEvent As Boolean = PaintEventEnabled
        Dim oldEnabled As Boolean = Me.Enabled
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------- save position 
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        ' -----------------------------------------------------
        RTB_Lines = Me.Lines
        ColorLine(0, Color.Black, Color.LightYellow)
        ColorLine(1, Color.Black, Color.LightYellow)
        ' -----------------------------------------------------
        For i As Int32 = 2 To RTB_Lines.Length - 1
            Dim s As String = ReplaceMultipleSpacesAndTabAndTrim(RTB_Lines(i))
            s = s.Replace(" ", vbTab)
            If s.ToUpper.StartsWith("X") Then
                If s.Split(CChar(vbTab)).Length = 7 Then
                    ColorLine(i, Color.Black, Color.FromArgb(240, 255, 200))
                Else
                    ColorLine(i, Color.Black, Color.Orange)
                End If
                XAxisAutoIfTime(i)
                ReplaceKeywords(New String() {"X", "Time"}, i)
            ElseIf s.ToUpper.StartsWith("Y") Then
                If s.Split(CChar(vbTab)).Length = 7 Then
                    ColorLine(i, Color.Black, Color.FromArgb(240, 255, 200))
                Else
                    ColorLine(i, Color.Black, Color.Orange)
                End If
                ReplaceKeywords(New String() {"Y"}, i)
            ElseIf s.ToUpper.StartsWith("LEFTCOLUMNSCOUNT") Or s.ToUpper.StartsWith("COLUMNSIZE") Or s.ToUpper.StartsWith("NUMDIGITS") Then
                If s.Split(CChar(vbTab)).Length = 2 Then
                    ColorLine(i, Color.Black, Color.FromArgb(240, 255, 200))
                Else
                    ColorLine(i, Color.Black, Color.Orange)
                End If
                ReplaceKeywords(New String() {"LeftColumnsCount", "ColumnSize", "NumDigits"}, i)
            Else
                ColorLine(i, Color.Black, Color.FromArgb(240, 240, 240))
            End If
        Next
        ' ---------------------------------------------------------- restore
        RestoreSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        PaintEventEnabled = oldPaintEvent
        Me.Enabled = oldEnabled
        Me.Focus()
    End Sub

    Private Sub XAxisAutoIfTime(ByVal i As Int32)
        If RTB_Lines(i).ToUpper.StartsWith("X" + vbTab + "TIME") Then NormalizeXAxisForTime(i)
        If RTB_Lines(i).ToUpper.StartsWith("X" + vbTab + "TSEC") Then NormalizeXAxisForTime(i)
        If RTB_Lines(i).ToUpper.StartsWith("X" + vbTab + "SEC") Then NormalizeXAxisForTime(i)
        If RTB_Lines(i).ToUpper.StartsWith("X" + vbTab + "TMIN") Then NormalizeXAxisForTime(i)
        If RTB_Lines(i).ToUpper.StartsWith("X" + vbTab + "MIN") Then NormalizeXAxisForTime(i)
        If RTB_Lines(i).ToUpper.StartsWith("X" + vbTab + "THRS") Then NormalizeXAxisForTime(i)
        If RTB_Lines(i).ToUpper.StartsWith("X" + vbTab + "HOURS") Then NormalizeXAxisForTime(i)
        If RTB_Lines(i).ToUpper.StartsWith("X" + vbTab + "TDAY") Then NormalizeXAxisForTime(i)
        If RTB_Lines(i).ToUpper.StartsWith("X" + vbTab + "DAYS") Then NormalizeXAxisForTime(i)
    End Sub

    Private Sub NormalizeXAxisForTime(ByVal i As Int32)
        Dim s As String = "X" + vbTab + "TIME" + vbTab + "0" + vbTab + "-" + vbTab + "-" + vbTab + "-" + vbTab + "BLACK"
        If RTB_Lines(i) <> s Then
            ' ---------------------------------------------------------- save position 
            Dim OldCursorPosition As Integer = SelectionStart
            ' ---------------------------------------------------------- set the whole line colors
            SelectLine(i)
            Me.SelectedText = s
            ' ---------------------------------------------------------- restore
            SelectionStart = OldCursorPosition
            SelectionLength = 0
        End If
    End Sub


    Private Sub ColorLine(ByVal i As Int32, ByVal front As Color, ByVal back As Color)
        ' ---------------------------------------------------------- save position 
        Dim OldCursorPosition As Integer = SelectionStart
        ' ---------------------------------------------------------- set the whole line colors
        SelectLine(i)
        SelectionColor = front
        SelectionBackColor = back
        ' ---------------------------------------------------------- restore
        SelectionStart = OldCursorPosition
        SelectionLength = 0
    End Sub

    Friend Sub SelectLine(ByVal line As Int32)
        If line >= 0 AndAlso line < Lines.Length AndAlso Lines.Length > 0 Then
            Dim lineStart As Int32 = GetFirstCharIndexFromLine(line)
            Dim lineLength As Int32 = Lines(line).Length
            Me.Select(lineStart, lineLength)
        Else
            Me.Select(0, 0)
        End If
    End Sub

    ' ===============================================================================================================
    '   REPLACE KEYWORDS
    ' ===============================================================================================================
    Private Sub ReplaceKeywords(ByVal Keywords As String(), _
                                ByVal lineNumber As Int32)
        '
        Dim line As String = RTB_Lines(lineNumber)
        Dim LineStart As Int32
        If lineNumber >= 0 Then LineStart = GetFirstCharIndexFromLine(lineNumber)
        '
        Dim position As Int32
        For Each kw As String In Keywords
            If kw = "" Then Continue For
            position = 0
            Do
                position = line.IndexOf(kw, position, StringComparison.InvariantCultureIgnoreCase)
                If position >= 0 Then
                    If IsCompleteWord(line, position, kw.Length) Then
                        If lineNumber >= 0 Then
                            Me.Select(LineStart + position, kw.Length)
                            If kw <> Me.SelectedText Then
                                Me.SelectedText = kw
                            End If
                        Else
                            Mid(line, position + 1, kw.Length) = kw
                        End If
                    End If
                    position += kw.Length
                End If
            Loop Until position < 0
        Next
    End Sub

    Private Function IsCompleteWord(ByVal l As String, ByVal position As Int32, ByVal length As Int32) As Boolean
        ' ------------------------------------------------------------ test left side
        If position > 0 AndAlso _
           IsLetterOrSpaceOrNumber(l(position - 1)) Then Return False
        ' ------------------------------------------------------------- test right side
        If position + length < l.Length AndAlso _
           IsLetterOrSpaceOrNumber(l(position + length)) Then Return False
        ' -------------------------------------------------------------
        Return True
    End Function

    Private Function IsLetterOrSpaceOrNumber(ByVal c As Char) As Boolean
        Return (c >= "a" And c <= "z") OrElse (c >= "A" And c <= "Z") OrElse (c >= "0" And c <= "9") OrElse c = "_"
    End Function

    'Private Function InsideQuotes(ByVal index As Int32, ByVal text As String) As Boolean
    '    InsideQuotes = False
    '    If index >= text.Length Then Return False
    '    For Counter As Int32 = 0 To index - 1
    '        If text(Counter) = """"c Then
    '            InsideQuotes = Not InsideQuotes
    '        End If
    '    Next
    'End Function


    ' ===============================================================================================================
    '   RTB - SAVE and RESTORE selection and scroll ( scroll aligned with the first text line )
    ' ===============================================================================================================
    Friend Structure ScrollData
        Dim OldSelectionStart As Int32
        Dim OldSelectionLength As Int32
        Dim OldSelectionFirstChar As Int32
        Dim OldSelectionLastChar As Int32
        Dim OldSelectionFirstLine As Int32
        Dim OldSelectionLastLine As Int32
    End Structure

    Friend Sub SaveSelectionAndScroll(ByRef sd As ScrollData)
        sd.OldSelectionStart = Me.SelectionStart
        sd.OldSelectionLength = Me.SelectionLength
        sd.OldSelectionFirstChar = Me.GetCharIndexFromPosition(New Point(1, 1))
        sd.OldSelectionLastChar = Me.GetCharIndexFromPosition(New Point(1, Me.ClientSize.Height))
        sd.OldSelectionFirstLine = Me.GetFirstSelectedLine
        sd.OldSelectionLastLine = Me.GetLastSelectedLine
    End Sub

    Friend Sub RestoreSelectionAndScroll(ByRef sd As ScrollData)
        '
        ' ---------------------------------------------------------- restore scroll 
        Me.Select(sd.OldSelectionFirstChar, 0)
        Me.ScrollToCaret()
        '
        ' ---------------------------------------------------------- restore selection point (if ...)
        If sd.OldSelectionStart >= sd.OldSelectionFirstChar Then
            Me.Select(sd.OldSelectionStart, sd.OldSelectionLength)
        End If
    End Sub

    Friend Sub RestoreScroll(ByRef sd As ScrollData)
        Me.Select(sd.OldSelectionFirstChar, 0)
        Me.ScrollToCaret()
    End Sub

    Friend Sub RestoreSelectedLinesAndScroll(ByRef sd As ScrollData)
        Me.Select(sd.OldSelectionFirstChar, 0)
        Me.ScrollToCaret()
        ' --------------------------------------------------------------
        If sd.OldSelectionLastLine >= sd.OldSelectionFirstLine Then
            Dim firstch As Int32 = GetFirstCharIndexFromLine(sd.OldSelectionFirstLine)
            Dim lastch As Int32 = GetLastCharIndexFromLine(sd.OldSelectionLastLine)
            Me.Select(firstch, lastch - firstch)
        End If
    End Sub

    Friend Function GetLastCharIndexFromLine(ByVal nline As Int32) As Int32
        If RTB_Lines.Length = 0 Then Return 0
        Return Me.GetFirstCharIndexFromLine(nline) + RTB_Lines(nline).Length
    End Function

    Friend Function GetFirstSelectedLine() As Int32
        Return Me.GetLineFromCharIndex(Me.SelectionStart)
    End Function
    Private Function GetLastSelectedLine() As Int32
        Return Me.GetLineFromCharIndex(Me.SelectionStart + Me.SelectionLength)
    End Function

    Friend Sub SelectCompleteWord_Short()
        Dim start, length As Int32
        FindWordPointers_ShortWords(Me.Text, Me.SelectionStart, start, length)
        Me.SelectionStart = start
        Me.SelectionLength = length
    End Sub

    Friend Sub SelectCompleteWord_Long()
        Dim start, length As Int32
        FindWordPointers(Me.Text, Me.SelectionStart, start, length)
        Me.SelectionStart = start
        Me.SelectionLength = length
    End Sub


    ' ===================================================================================
    '  DOUBLE CLICK
    ' ===================================================================================
    Private Sub SimpleRTB_DoubleClick(ByVal sender As Object, ByVal e As EventArgs) Handles Me.DoubleClick
        SelectCompleteWord_Long()
    End Sub

    ' ===============================================================================================
    '  ENABLE DISABLE PAINT
    ' ===============================================================================================
    Friend PaintEventEnabled As Boolean = True

    <DebuggerStepThrough()> _
    Protected Overrides Sub WndProc(ByRef m As Message)
        Const WM_PAINT As Int32 = &HF
        If m.Msg = WM_PAINT AndAlso Not PaintEventEnabled Then
            m.Result = IntPtr.Zero
        Else
            MyBase.WndProc(m)
        End If
    End Sub

    ' ===============================================================================================
    '   SAVE/RESTORE the SCROLL POS
    ' ===============================================================================================
    Private Declare Auto Function RtfScroll _
                Lib "user32.dll" Alias "SendMessage" (ByVal hWnd As IntPtr, _
                                                      ByVal Msg As Integer, _
                                                      ByVal wParam As IntPtr, _
                                                      ByRef lParam As System.Drawing.Point) As Integer

    Private Declare Function LockWindowUpdate Lib "user32" (ByVal hWnd As Integer) As Integer

    Private Const WM_USER As Int32 = &H400
    Private Const EM_GETSCROLLPOS As Int32 = WM_USER + 221
    Private Const EM_SETSCROLLPOS As Int32 = WM_USER + 222

    Private TextBoxOriginalScroll As Int32

    Private pt As New System.Drawing.Point()
    Friend Sub SaveScroll()
        If Me.Focused Then
            LockWindowUpdate(Me.Handle.ToInt32)
            TextBoxOriginalScroll = Me.SelectionStart
        Else
            RtfScroll(Me.Handle, EM_GETSCROLLPOS, IntPtr.Zero, pt)
        End If
    End Sub
    Friend Sub RestoreScroll()
        If Me.Focused Then
            Me.SelectionStart = TextBoxOriginalScroll
            Me.ScrollToCaret()
        Else
            RtfScroll(Me.Handle, EM_SETSCROLLPOS, IntPtr.Zero, pt)
        End If
        LockWindowUpdate(0)
    End Sub

    Friend Sub ScrollToCenterLine(ByVal line As Int32)
        Dim VisibleLinesCount As Int32 = GetLastVisibleLine() - GetFirstVisibleLine()
        Dim OldEventsAreEnabled As Boolean = EventsAreEnabled
        EventsAreEnabled = False
        ' ---------------------------------------------------------- 
        Dim oldPaintEvent As Boolean = PaintEventEnabled
        Dim oldEnabled As Boolean = Me.Enabled
        PaintEventEnabled = False
        Me.Enabled = False
        ' ----------------------------------------------------------
        Dim delta As Int32 = VisibleLinesCount \ 2 - 1
        ScrollToLine(line + delta)
        ScrollToLine(line - delta)
        ' ---------------------------------------------------------- 
        PaintEventEnabled = oldPaintEvent
        Me.Enabled = oldEnabled
        EventsAreEnabled = OldEventsAreEnabled
    End Sub

    Friend Sub ScrollToLine(ByVal nline As Int32)
        If RTB_Lines.Length = 0 Then Return
        If nline < 0 Then nline = 0
        If nline > RTB_Lines.Length - 1 Then nline = RTB_Lines.Length - 1
        Me.Select(GetFirstCharIndexFromLine(nline), 1)
        Me.ScrollToCaret()
    End Sub

    Friend Sub SelectLines(ByVal FirstLine As Int32, ByVal LastLine As Int32)
        If RTB_Lines.Length > 0 Then
            Dim FirstChar As Int32 = Me.GetFirstCharIndexFromLine(FirstLine)
            If LastLine > RTB_Lines.Length - 1 Then LastLine = RTB_Lines.Length - 1
            Dim LastChar As Int32 = Me.GetFirstCharIndexFromLine(LastLine) + RTB_Lines(LastLine).Length
            Me.Select(FirstChar, LastChar - FirstChar)
        End If
    End Sub

    Friend Sub SetFirstVisibleLine(ByVal nline As Int32)
        ScrollToLine(nline)
        Me.SelectionLength = 0
    End Sub
    Friend Function GetFirstVisibleLine() As Int32
        Return Me.GetLineFromCharIndex(Me.GetCharIndexFromPosition(New Point(1, 1)))
    End Function
    Friend Function GetLastVisibleLine() As Int32
        Dim p As New Point(1, Me.ClientSize.Height)
        Return Me.GetLineFromCharIndex(Me.GetCharIndexFromPosition(p))
    End Function

    Friend Sub ScrollToLastLine()
        Me.SelectionLength = 0
        Me.SelectionStart = Me.Text.Length
        Me.ScrollToCaret()
    End Sub

    Friend Sub ScrollToFirstLine()
        Me.SelectionLength = 0
        Me.SelectionStart = 0
        Me.ScrollToCaret()
    End Sub

    ' ===============================================================================================
    '   SUPPRESS SOUNDS
    ' ===============================================================================================
    Friend Sub RichTextBox_SuppressSounds(ByVal sender As Object, ByRef e As KeyEventArgs)
        Dim rtb As RichTextBox = CType(sender, RichTextBox)
        If rtb.GetLineFromCharIndex(rtb.SelectionStart) = 0 AndAlso _
                                            (e.KeyData = Keys.Up Or e.KeyData = Keys.PageUp) _
        OrElse rtb.GetLineFromCharIndex(rtb.SelectionStart) = rtb.GetLineFromCharIndex(rtb.TextLength) AndAlso _
                                            (e.KeyData = Keys.Down Or e.KeyData = Keys.PageDown) _
        OrElse rtb.SelectionStart = rtb.TextLength AndAlso e.KeyData = Keys.Right _
        OrElse rtb.SelectionStart = 0 AndAlso _
                (e.KeyData = Keys.Left Or (e.KeyData = Keys.Back And rtb.SelectionLength = 0)) Then
            e.Handled = True
        End If
    End Sub

    ' ===============================================================================================
    '  CUT COPY and PASTE
    ' ===============================================================================================
    Friend Sub ExecCut()
        If Me.SelectionLength > 0 Then
            SendTextToClipboard()
            ExecDelete()
        End If
    End Sub
    Friend Sub ExecCopy()
        If Me.SelectionLength > 0 Then
            SendTextToClipboard()
        End If
    End Sub
    Private Sub SendTextToClipboard()
        ' ------------------------------------------------------- clipboard text must have the CRLF
        Dim s As String = Me.SelectedText.Replace(vbLf, vbCrLf)
        ' ------------------------------------------------------- send to clipboard
        Dim data As IDataObject
        data = New DataObject(DataFormats.Text, s)
        Try
            Clipboard.SetDataObject(data, True)
        Catch
        End Try
    End Sub

    Friend Sub ExecPaste()
        ' ----------------------------------------------------------------
        'Dim sw As Stopwatch = New Stopwatch
        'sw.Start()
        ' ----------------------------------------------------------------
        If Clipboard.GetDataObject().GetDataPresent(DataFormats.Text) Then
            ' ------------------------------------------------------------ gray backcolor
            Dim oldcolor As Color = BackColor
            BackColor = Color.FromArgb(255, 230, 190)
            Refresh()
            ' ------------------------------------------------------------ wait cursor
            Cursor.Current = Cursors.WaitCursor
            Refresh()
            Application.DoEvents()
            WaitKeyOff(Keys.V)
            ' ------------------------------------------------------------ PaintEvent and Event (BUT NOT Enabled)
            Dim oldEventsAreEnabled As Boolean = EventsAreEnabled
            Dim oldPaintEvent As Boolean = PaintEventEnabled
            PaintEventEnabled = False
            EventsAreEnabled = False
            ' ------------------------------------------------------------ correct CRLF from clipbboard to single LF
            Dim TextToPaste As String = Clipboard.GetText.Replace(vbCrLf, vbLf)
            '
            ' ------------------------------------------------------------ paste the text and show immediately
            PaintEventEnabled = True
            EventsAreEnabled = True
            Me.SelectedText = TextToPaste
            RTB_Lines = Me.Lines
            PaintEventEnabled = False
            EventsAreEnabled = False
            ' ------------------------------------------------------------ prepare EndLine
            Dim EndLine As Int32 = GetLineFromCharIndex(SelectionStart + SelectionLength)
            '
            ' ------------------------------------------------------------ set the final cursor position
            'Me.SelectionLength = 0
            'Dim i As Int32 = GetLastCharIndexFromLine(EndLine) - DistanceToEndOfLine
            'If i < 0 Then i = 0
            'Me.SelectionStart = i
            ' ------------------------------------------------------------ 
            Me.ReformatText()
            Me.ColorLines()
            ' ------------------------------------------------------------ 
            PaintEventEnabled = oldPaintEvent
            EventsAreEnabled = oldEventsAreEnabled
            Me.Enabled = True
            ' ------------------------------------------------------------
            Me.Focus()
            Cursor.Current = Cursors.Default
            ' ------------------------------------------------------------ restore the original backcolor
            BackColor = oldcolor
            Refresh()
        End If
        ' -----------------------------------------------------------------
        'Form1.Text = sw.ElapsedMilliseconds.ToString
    End Sub

    Friend Sub ExecDelete()
        If Me.SelectionLength > 0 Then
            ' ------------------------------------------------------------
            Dim oldSelectionStart As Int32 = SelectionStart
            Dim startline As Int32 = GetLineFromCharIndex(SelectionStart)
            ' ------------------------------------------------------------ 
            Dim oldPaintEvent As Boolean = PaintEventEnabled
            Dim oldEnabled As Boolean = Me.Enabled
            PaintEventEnabled = False
            Me.Enabled = False
            ' ------------------------------------------------------------ 
            PaintEventEnabled = True
            Me.Enabled = True
            Me.SelectedText = ""
            RTB_Lines = Me.Lines
            PaintEventEnabled = False
            Me.Enabled = False
            ' ------------------------------------------------------------
            Me.SelectionStart = oldSelectionStart
            Me.SelectionLength = 0
            ' ------------------------------------------------------------
            PaintEventEnabled = oldPaintEvent
            Me.Enabled = oldEnabled
            ' ------------------------------------------------------------
            Me.Focus()
            ' ------------------------------------------------------------
            ScrollToCaret_if_DeletingLastLines()
        End If
    End Sub

    ' This enable to scroll immediately when deleting many lines at the end of the file
    ' (test with the file ERRORS_EDIT-1.txt)
    Private Sub ScrollToCaret_if_DeletingLastLines()
        If Me.GetPositionFromCharIndex(Me.Text.Length).Y < Me.ClientSize.Height - 32 Then
            Me.ScrollToCaret()
        End If
    End Sub

    ' ===============================================================================================
    '  COMMENT UN-COMMENT
    ' ===============================================================================================
    Friend Sub CommentSelectedLines(ByVal comment As Boolean)
        ' ---------------------------------------------------------------------
        PushUndoState()
        ' ---------------------------------------------------------------------
        RTB_Lines = Me.Lines
        If RTB_Lines.Length > 0 Then
            ' -----------------------------------------------------------------
            Dim oldPaintEvent As Boolean = PaintEventEnabled
            Dim oldEnabled As Boolean = Me.Enabled
            PaintEventEnabled = False
            Me.Enabled = False
            ' -----------------------------------------------------------------
            Dim FirstSelectedLine As Int32 = GetFirstSelectedLine()
            Dim LastSelectedLine As Int32 = GetLastSelectedLine()
            ' -----------------------------------------------------------------
            Dim CommentColumn As Int32 = 999999
            For i As Int32 = FirstSelectedLine To LastSelectedLine
                Dim lineText As String = RTB_Lines(i)
                Dim firstch As Int32 = CountInitialChars(lineText)
                If firstch < CommentColumn Then CommentColumn = firstch
            Next
            ' -----------------------------------------------------------------
            For i As Int32 = FirstSelectedLine To LastSelectedLine
                Dim lineText As String = RTB_Lines(i)
                Dim firstch As Int32 = CountInitialChars(lineText)
                If firstch >= 0 Then
                    Dim start As Int32 = Me.GetFirstCharIndexFromLine(i)
                    If comment Then
                        Me.Select(start + CommentColumn, 0)
                        Me.SelectedText = "'"
                    Else
                        Me.Select(start + firstch, 1)
                        If Me.SelectedText = "'" Then
                            Me.SelectedText = ""
                        End If
                    End If
                End If
            Next
            ' -----------------------------------------------------------------
            PaintEventEnabled = oldPaintEvent
            Me.Enabled = oldEnabled
            ' -----------------------------------------------------------------
            Me.SelectLines(FirstSelectedLine, LastSelectedLine)
        End If
        Me.Focus()
    End Sub

    Private Function CountInitialChars(ByVal s As String) As Int32
        Dim CharCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : CharCount += 1
                Case " " : CharCount += 1
                Case Else : Exit For
            End Select
        Next
        Return CharCount
    End Function

    Private Function CountInitialTabs(ByVal s As String) As Int32
        Dim TabCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : TabCount += 1
                Case Else : Exit For
            End Select
        Next
        Return TabCount
    End Function

    Private Function CountInitialTabs_NotAllowingSpaces(ByVal s As String) As Int32
        Dim TabCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : TabCount += 1
                Case " " : Return -9999           ' Spaces are not allowed
                Case Else : Exit For
            End Select
        Next
        Return TabCount
    End Function


    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################
    ' 
    '   UNDO
    ' 
    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################

    Public Structure TextState
        Public Text As String
        Public CursorPosition As Integer
        Public CursorLength As Integer
    End Structure
    Private UndoStack As New Stack(Of TextState)
    Private RedoStack As New Stack(Of TextState)

    Friend Function Editor_CanUndo() As Boolean
        Return UndoStack.Count > 0
    End Function
    Friend Function Editor_CanRedo() As Boolean
        Return RedoStack.Count > 0
    End Function

    Public Function GetTextState() As TextState
        GetTextState.Text = Me.Text
        GetTextState.CursorPosition = Me.SelectionStart
        GetTextState.CursorLength = Me.SelectionLength
    End Function

    Public Sub SetTextState(ByVal ts As TextState)
        Me.Enabled = False
        PaintEventEnabled = False
        '
        Me.Text = ts.Text
        Me.SelectionStart = ts.CursorPosition
        Me.SelectionLength = ts.CursorLength
        '
        PaintEventEnabled = True
        Me.Enabled = True
    End Sub

    Friend Sub PushUndoState()
        UndoStack.Push(GetTextState)
        RedoStack.Clear()
    End Sub

    ' ===============================================================================================
    '  UNDO and REDO commands
    ' ===============================================================================================
    Friend Sub ExecUndo()
        If Editor_CanUndo() Then
            RedoStack.Push(GetTextState)
            SetTextState(UndoStack.Pop())
        End If
    End Sub
    Friend Sub ExecRedo()
        If Editor_CanRedo() Then
            UndoStack.Push(GetTextState)
            SetTextState(RedoStack.Pop())
        End If
    End Sub

    ' ===============================================================================================
    '  KEY AND MOUSE EVENTS
    ' ===============================================================================================
    Private Sub Me_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyDown
        '
        If My.Computer.Keyboard.CtrlKeyDown Then
            ' ------------------------------------------------------- CUT / COPY / PASTE / UNDO / REDO
            RTB_Lines = Me.Lines
            e.SuppressKeyPress = True
            e.Handled = True
            Select Case e.KeyCode
                'Case Keys.F : PushUndoState(): Form1.Editor_Find()
                'Case Keys.R : PushUndoState(): Form1.Editor_FindReplace()
                Case Keys.X : PushUndoState() : Me.ExecCut()
                Case Keys.C : Me.ExecCopy()
                Case Keys.V : PushUndoState() : Me.ExecPaste()
                Case Keys.Z : Me.ExecUndo()
                Case Keys.Y : Me.ExecRedo()
                Case Else
                    e.Handled = False
                    e.SuppressKeyPress = False
            End Select
            Me.Focus()
        Else
            ' ------------------------------------------------------- PUSH UNDO STATE
            Select Case e.KeyCode
                Case Keys.Up, Keys.Down, Keys.Left, Keys.Right
                Case Keys.Insert
                    If e.Modifiers <> Keys.None Then
                        e.Handled = True
                    Else
                        OverwriteMode_Toggle()
                    End If
                Case Else
                    PushUndoState()
            End Select
        End If
        ' ----------------------------------------------------------- Suppress RTB sounds
        RichTextBox_SuppressSounds(sender, e)
    End Sub

    Private Sub SimpleRTB_KeyUp(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyUp
        OverwriteMode_ShowCaret()
    End Sub

    Private Sub SimpleRTB_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        OverwriteMode_ShowCaret()
    End Sub

    ' ===============================================================================================
    '   TEXT CHANGED
    ' ===============================================================================================
    Private Sub SimpleRTB_TextChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.TextChanged
        If Not EventsAreEnabled Then Return
        ReformatText()
        ColorLines()
        Refresh()
    End Sub

    ' ===============================================================================================
    '   SELECTION CHANGED
    ' ===============================================================================================
    Private Sub SimpleRTB_SelectionChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.SelectionChanged
        '
        ' --------------------------------------------------------------------- AUTO WORD SELECTION BUG CORRECTION 
        Me.AutoWordSelection = True
        Me.AutoWordSelection = False
        ' ---------------------------------------------------------------------
    End Sub

    ' ======================================================================================================
    '  INSERT - OVERWRITE MODE - BLACK SQUARE CARET
    ' ======================================================================================================
    Private Declare Function CreateCaret Lib "user32" (ByVal hWnd As IntPtr, _
                                                      ByVal hBitmap As IntPtr, _
                                                      ByVal nWidth As Int32, _
                                                      ByVal nHeight As Int32) As Int32

    Private Declare Function ShowCaret Lib "user32" (ByVal hWnd As IntPtr) As Int32

    Private OverwriteMode As Boolean

    Friend Sub OverwriteMode_Toggle()
        OverwriteMode = Not OverwriteMode
        If Not OverwriteMode Then
            CreateCaret(Me.Handle, IntPtr.Zero, 1, CInt((Me.Font.Height - 1) * Me.ZoomFactor))
            ShowCaret(Me.Handle)
        End If
    End Sub

    Friend Sub OverwriteMode_ShowCaret()
        If OverwriteMode Then
            CreateCaret(Me.Handle, IntPtr.Zero, CInt(12 * Me.ZoomFactor), CInt((Me.Font.Height - 1) * Me.ZoomFactor))
            ShowCaret(Me.Handle)
        End If
    End Sub

End Class
