﻿Module Module_Options

    Friend Structure AxisDefinition
        Dim Axis As String
        Dim Name As String
        Dim Column As Int32
        Dim Min As Double
        Dim Max As Double
        Dim Stp As Double
        Dim ValueMin As Double
        Dim ValueMax As Double
        Dim Smoothing As Double
        Dim OldValue As Double
        Dim PosX As Single
        Dim _Color As Color
        Dim _Pen As Pen
        Dim _Brush As SolidBrush
    End Structure

    Friend Axes(-1) As AxisDefinition
    Friend XaxisIsTime As Boolean
    Friend XaxisTimeUnit As Int32

    Private Sub DecodeOptions()
        'On Error Resume Next
        XaxisIsTime = False
        XaxisTimeUnit = 1
        ColumnSize = 44
        LeftColumnsCount = 0
        ReDim Axes(-1)
        Dim lines() As String = Form2.TextBox1.Lines
        For n As Int32 = 0 To lines.Length - 1
            Dim l1 As String = lines(n)
            l1 = RemoveComments(l1)
            l1 = ReplaceMultipleSpacesAndTabAndTrim(l1)
            Dim l2 As String = l1.ToUpper
            If l2.StartsWith("END") Then Exit For
            ' --------------------------------------------------------------------------
            Dim sa() As String = l1.Split(" "c)
            If sa.Length = 7 Then
                Dim sa0 As String = sa(0).ToUpper
                Dim sa1 As String = sa(1).ToUpper
                '
                If sa0.StartsWith("X") Or sa0.StartsWith("Y") Then
                    ReDim Preserve Axes(Axes.Length)
                    With Axes(Axes.Length - 1)
                        .Axis = sa(0)
                        .Name = sa(1)
                        .Column = CInt(Val(sa(2)))
                        If Information.IsNumeric(sa(3)) Then
                            .Min = Val(sa(3))
                        Else
                            .Min = Double.NaN
                        End If
                        If Information.IsNumeric(sa(4)) Then
                            .Max = Val(sa(4))
                        Else
                            .Max = Double.NaN
                        End If
                        .Stp = Val(sa(5))
                        ._Color = DecodeColor(sa(6))
                        ._Pen = New Pen(._Color)
                        ._Brush = New SolidBrush(._Color)
                    End With
                End If
                ' -------------------------------------------------------------------------- test if X-Axis is Time
                If sa0.StartsWith("X") Then
                    ' TODO ----- XaxisIsTime Options
                    If sa1.Contains("TIME") Then XaxisIsTime = True
                    SetXaxisTimeUnit()
                End If
            Else
                'Beep()
            End If
        Next
        ' ---------------------------------------------------------
        DecodeSpecialOptions()
        SetSmoothing()
    End Sub

    Friend Sub SetXaxisTimeUnit()
        If Axes.Length = 0 Then Return
        Dim seconds As Double = Axes(0).ValueMax - Axes(0).ValueMin
        If seconds > 3600 * 24 Then
            XaxisTimeUnit = 3600 * 24
            Axes(0).Name = "Days"
        ElseIf seconds > 3600 Then
            XaxisTimeUnit = 3600
            Axes(0).Name = "Hours"
        ElseIf seconds > 60 Then
            XaxisTimeUnit = 60
            Axes(0).Name = "Minutes"
        Else
            XaxisTimeUnit = 1
            Axes(0).Name = "Seconds"
        End If
    End Sub

    Private Function DecodeColor(ByVal s As String) As Color
        DecodeColor = Color.FromArgb(0, 0, 0, 0)
        DecodeColor = Color.FromName(s)
        If DecodeColor.A = 0 Then
            s = s.Replace(";", ",")
            Dim sa() As String = s.Split(","c)
            If sa.Length = 3 Then
                DecodeColor = Color.FromArgb(Val_Int_0_255(sa(0)), _
                                             Val_Int_0_255(sa(1)), _
                                             Val_Int_0_255(sa(2)))
            Else
                DecodeColor = Color.Black
            End If
        End If
    End Function

    Private Function Val_Int_0_255(ByVal l As String) As Int32
        Dim n As Double = Val(l)
        If n < 0 Then n = 0
        If n > 255 Then n = 255
        Return CInt(n)
    End Function

    Private Sub DecodeSpecialOptions()
        'On Error Resume Next
        Dim lines() As String = Form2.TextBox1.Lines
        For n As Int32 = 0 To lines.Length - 1
            Dim l1 As String = lines(n)
            l1 = RemoveComments(l1)
            l1 = ReplaceMultipleSpacesAndTabAndTrim(l1)
            Dim l2 As String = l1.ToUpper
            ' ----------------------------------------------
            Dim sa() As String = l1.Split(" "c)
            If sa.Length = 2 Then
                Dim sa0 As String = sa(0).ToUpper
                If sa0.StartsWith("LEFTCOLUMNSCOUNT") Then
                    LeftColumnsCount = CInt(Val(sa(1)))
                    If LeftColumnsCount > Axes.Length - 1 Then LeftColumnsCount = Axes.Length - 1
                End If
                If sa0.StartsWith("COLUMNSIZE") Then
                    ColumnSize = CInt(Val(sa(1)))
                    If ColumnSize < 24 Then ColumnSize = 24
                    If ColumnSize > 99 Then ColumnSize = 99
                End If
                If sa0.StartsWith("NUMDIGITS") Then
                    NumDigits = CInt(Val(sa(1)))
                    If NumDigits < 2 Then NumDigits = 2
                    If NumDigits > 9 Then NumDigits = 9
                End If
            ElseIf sa.Length = 3 Then
                'Dim sa0 As String = sa(0).ToUpper
                'Dim sa1 As String = sa(1).ToUpper
                'If sa0.StartsWith("SMOOTHING") Then
                '    For i As Int32 = 0 To Axes.Length - 1
                '        If sa1 = Axes(i).Name.ToUpper Then
                '            Axes(i).Smoothing = Math.Max(Val(sa(2)), 1)
                '        End If
                '    Next
                'End If
            Else
                'Beep()
            End If
        Next
    End Sub

    Friend Sub SetSmoothing()
        Dim n As Int32
        n = Form1.TrackBar2.Value
        For i As Int32 = 1 To LeftColumnsCount
            Axes(i).Smoothing = n
        Next
        n = Form1.TrackBar3.Value
        For i As Int32 = LeftColumnsCount + 1 To Axes.Length - 1
            Axes(i).Smoothing = n
        Next
    End Sub

    ' ==============================================================
    '  OPTIONS FILE
    ' ==============================================================
    Friend OptionsFilePathAndName As String = Application.StartupPath

    Friend Sub ReadOptionsFile()
        If OptionsFilePathAndName = Nothing Then Return
        If Not IO.File.Exists(OptionsFilePathAndName) Then Return
        Dim old As Boolean = EventsAreEnabled
        EventsAreEnabled = False
        Form2.TextBox1.InitWithText(IO.File.ReadAllText(OptionsFilePathAndName))
        Form2.TextBox1.SetFirstVisibleLine(Form2.Options_FirstVisibleLine)
        EventsAreEnabled = True
        DecodeOptions()
        Form2.GroupBox1.Text = IO.Path.GetFileName(OptionsFilePathAndName)
    End Sub

    Friend Sub WriteOptionsFile()
        IO.File.WriteAllText(OptionsFilePathAndName, Form2.TextBox1.Text)
        DecodeOptions()
    End Sub

End Module
