﻿
Module Module_SaveLoad

    Friend EventsAreEnabled As Boolean = False

    Friend Form2_IsVisible As Boolean
    Friend Form2_VisibleAtStart As Boolean
    Friend Form2_PositionX As Int32
    Friend Form2_PositionY As Int32

    Friend ZoomFactor_Form2 As Single

    Friend GCI As Globalization.CultureInfo = Globalization.CultureInfo.InvariantCulture

    ' =======================================================================================================
    '   APP TITLE AND VERSION (ALL FIELDS)
    ' =======================================================================================================
    Friend Function AppTitleAndVersion(Optional ByVal RemoveThereminoPrefix As Boolean = False) As String
        Dim appname As String = IO.Path.GetFileNameWithoutExtension(Application.ExecutablePath)
        Dim Title As String = Replace(appname, "_", " ")
        If RemoveThereminoPrefix Then Title = Replace(Title, "Theremino ", "")
        Return Title + " - V" + GetVersionString()
    End Function

    ' =======================================================================================================
    '   GET VERSION (ALSO TRIPLE OR QUADRUPLE FIELDS)
    ' =======================================================================================================
    Friend Function GetVersionString() As String
        Dim s() As String = Split(My.Application.Info.Version.ToString, ".")
        GetVersionString = s(0) + "." + s(1)
        If s(2) <> "0" Then
            GetVersionString += "." + s(2)
            If s(3) <> "0" Then
                GetVersionString += "." + s(3)
            End If
        End If
    End Function

    ' =======================================================================================
    '  FORM FUNCTIONS
    ' =======================================================================================
    Friend Sub LimitFormPosition_CompletelyVisible(ByVal f As System.Windows.Forms.Form)
        If f.WindowState <> FormWindowState.Normal Then Return
        GetMaxScreenBounds()
        EnsureFormCompletelyVisible(f)
    End Sub

    Friend Sub LimitFormPosition(ByVal f As System.Windows.Forms.Form)
        If f.WindowState <> FormWindowState.Normal Then Return
        GetMaxScreenBounds()
        EnsureFormVisible(f)
    End Sub

    Private SB As Rectangle = New Rectangle(Integer.MaxValue, Integer.MaxValue, Integer.MinValue, Integer.MinValue)

    Private Sub GetMaxScreenBounds()
        For Each s As Screen In System.Windows.Forms.Screen.AllScreens
            SB = Rectangle.Union(SB, s.WorkingArea)
        Next
    End Sub

    Private Sub EnsureFormCompletelyVisible(ByVal frm As Form)
        With frm
            .Width = Math.Min(.Width, SB.Width)         ' not more than a maximized window
            .Height = Math.Min(.Height, SB.Height)      ' not more than a maximized window
            .Width = Math.Max(.Width, 32)               ' at least 32x24
            .Height = Math.Max(.Height, 24)             ' at least 32x24
            .Left = Math.Min(.Left, SB.Right - .Width)  ' not beyond the right border
            .Top = Math.Min(.Top, SB.Bottom - .Height)  ' not beyond the bottom border
            .Left = Math.Max(.Left, SB.Left)            ' at least at the left border
            .Top = Math.Max(.Top, SB.Top)               ' at least at the top border
        End With
    End Sub

    Private Sub EnsureFormVisible(ByVal frm As Form)
        With frm
            .Width = Math.Min(.Width, SB.Width)             ' not more than VIRTUALSCREEN dimensions
            .Height = Math.Min(.Height, SB.Height)          ' not more than VIRTUALSCREEN dimensions 
            .Width = Math.Max(.Width, 32)                   ' at least 32x24
            .Height = Math.Max(.Height, 24)                 ' at least 32x24
            .Left = Math.Min(.Left, SB.Right - 200)         ' not beyond right border - 200 pixels
            .Top = Math.Min(.Top, SB.Bottom - 80)          ' not beyond bottom border - 80 pixels
            .Left = Math.Max(.Left, SB.Left + 200 - .Width) ' at least at left border + 200 pixels
            .Top = Math.Max(.Top, SB.Top)                   ' at least at top border
        End With
    End Sub

    ' (The value of the RestoreBounds property is valid only 
    '   when the WindowState property of the Form class is not equal to Normal)
    Friend Function GetFormRectangle(ByVal frm As Form) As Rectangle
        Dim r As Rectangle
        If frm.WindowState = FormWindowState.Normal Then
            r = frm.Bounds
        Else
            r = frm.RestoreBounds
        End If
        Return r
    End Function


    ' ================================================================================================
    '  Private Read-Write functions
    ' ================================================================================================
    Private Function TabString(ByVal Name As String, _
                               Optional ByVal Value As Double = Double.NaN, _
                               Optional ByVal fmt As String = "") As String

        Dim nTab As Int32 = Math.Max(0, 22 - Name.Length)
        If Double.IsNaN(Value) Then
            Return Name
        Else
            Return Name & "=" & Strings.StrDup(nTab, " ") & Value.ToString(fmt, GCI)
        End If
    End Function
    Private Function TabString(ByVal Name As String, _
                                  ByVal Value As Boolean) As String

        Dim nTab As Int32 = Math.Max(0, 22 - Name.Length)

        Return Name & "=" & Strings.StrDup(nTab, " ") & Value.ToString
    End Function
    Private Function TabString(ByVal Name As String, _
                                ByVal Value As String) As String

        Dim nTab As Int32 = Math.Max(0, 22 - Name.Length)

        Return Name & "=" & Strings.StrDup(nTab, " ") & Value
    End Function
    Private Function Val_Double(ByVal l As String) As Double
        Return Val(l.Replace(",", "."))
    End Function
    Private Function Val_Single(ByVal l As String) As Single
        Return CSng(Val(l.Replace(",", ".")))
    End Function
    Private Function Val_Int(ByVal l As String) As Int32
        Return CInt(Val(l))
    End Function
    Private Function ExtractParamName(ByRef s As String) As String
        ' ------------------------- Returns the first field from begin to the first "=" symbol
        ' -------------------------  and removes it from the string
        Dim i As Int32
        i = InStr(s, "=")
        If i > 0 Then
            ExtractParamName = Trim(Strings.Left(s, i - 1))
            s = Trim(Mid(s, i + 1))
        Else
            ExtractParamName = Trim(s)
            s = ""
        End If
    End Function
    Private Function AssemblyName() As String
        Return System.Reflection.Assembly.GetExecutingAssembly.GetName.Name
    End Function


    ' ==================================================================================================
    '  SAVE LOAD -- Program INI
    ' ==================================================================================================
    Friend Sub Save_INI()
        Dim appname As String = IO.Path.GetFileNameWithoutExtension(Application.ExecutablePath)
        Dim iniFileName As String = Application.StartupPath + "\" + appname + "_INI.txt"
        Dim f As System.IO.StreamWriter = Nothing
        Try
            f = IO.File.CreateText(iniFileName)
            '
            f.WriteLine(" Program Params")
            f.WriteLine("===========================================")
            Dim r As Rectangle
            r = GetFormRectangle(Form1)
            f.WriteLine(TabString("Form1_Top", r.Top))
            f.WriteLine(TabString("Form1_Left", r.Left))
            f.WriteLine(TabString("Form1_Width", r.Width))
            f.WriteLine(TabString("Form1_Height", r.Height))
            f.WriteLine(TabString("Form1_WindowState", Form1.WindowState))
            '
            r = GetFormRectangle(Form2)
            f.WriteLine(TabString("Form2_Top", r.Top))
            f.WriteLine(TabString("Form2_Left", r.Left))
            f.WriteLine(TabString("Form2_Width", r.Width))
            f.WriteLine(TabString("Form2_Height", r.Height))
            'f.WriteLine(TabString("Form2_WindowState", Form2.WindowState))
            f.WriteLine(TabString("Form2_VisibleAtStart", Form2_IsVisible))
            f.WriteLine(TabString("Form2_Zoom", ZoomFactor_Form2))
            f.WriteLine(TabString("Form2_FirstLine", Form2.TextBox1.GetFirstVisibleLine))
            '
            f.WriteLine(TabString("ScrollBar1", Form1.TrackBar1.Value))
            f.WriteLine(TabString("ScrollBar2", Form1.TrackBar2.Value))
            f.WriteLine(TabString("ScrollBar3", Form1.TrackBar3.Value))
            '
            f.WriteLine(TabString("DataRelativePath", RelativePath_FromAbsolute(DataFilePathAndName)))
            f.WriteLine(TabString("OptionsRelativePath", RelativePath_FromAbsolute(OptionsFilePathAndName)))
            f.WriteLine(TabString("FastDrawing", Form1.FastDrawing))
            f.WriteLine(TabString("DrawLines", GraphPaper1.GraphType))
        Catch
        End Try
        Try
            f.Close()
        Catch
        End Try
    End Sub


    Friend Sub Load_INI()
        ' ------------------------------------------------------------- defaults
        '
        ' -------------------------------------------------------------
        ' ------------------------------------------------------------------------------- 
        ' With "Resume Next" subsequent parameters are loaded and f.Close() is executed
        ' -------------------------------------------------------------------------------
        On Error Resume Next  ' use Resume-Next instead of Try-Catch
        ' -------------------------------------------------------------------------------
        Dim l As String
        Dim appname As String = IO.Path.GetFileNameWithoutExtension(Application.ExecutablePath)
        Dim iniFileName As String = Application.StartupPath + "\" + appname + "_INI.txt"
        If My.Computer.FileSystem.FileExists(iniFileName) Then
            Dim f As System.IO.StreamReader
            f = IO.File.OpenText(iniFileName)
            Do While Not f.EndOfStream
                l = f.ReadLine()
                Select Case ExtractParamName(l)
                    Case "Form1_Top" : Form1.Top = Val_Int(l)
                    Case "Form1_Left" : Form1.Left = Val_Int(l)
                    Case "Form1_Width" : Form1.Width = Val_Int(l)
                    Case "Form1_Height" : Form1.Height = Val_Int(l)
                    Case "Form1_WindowState" : Form1.WindowState = CType((Val(l)), FormWindowState)
                        '
                    Case "Form2_Top" : Form2.Top = Val_Int(l) : Form2_PositionY = Val_Int(l)
                    Case "Form2_Left" : Form2.Left = Val_Int(l) : Form2_PositionX = Val_Int(l)
                    Case "Form2_Width" : Form2.Width = Val_Int(l)
                    Case "Form2_Height" : Form2.Height = Val_Int(l)
                        'Case "Form2_WindowState" : Form2.WindowState = CType((Val(l)), FormWindowState)
                    Case "Form2_VisibleAtStart" : Form2_VisibleAtStart = l = "True" : Form2_IsVisible = Form2_VisibleAtStart
                    Case "Form2_Zoom" : ZoomFactor_Form2 = Val_Single(l)
                    Case "Form2_FirstLine" : Form2.Options_FirstVisibleLine = Val_Int(l)
                        ' -----------------------------------------------------------------
                    Case "ScrollBar1" : Form1.TrackBar1.Value = Val_Int(l)
                    Case "ScrollBar2" : Form1.TrackBar2.Value = Val_Int(l)
                    Case "ScrollBar3" : Form1.TrackBar3.Value = Val_Int(l)
                        ' -----------------------------------------------------------------
                    Case "DataRelativePath" : DataFilePathAndName = AbsolutePath_FromRelative(l)
                    Case "OptionsRelativePath" : OptionsFilePathAndName = AbsolutePath_FromRelative(l)
                    Case "FastDrawing" : Form1.FastDrawing = l = "True"
                    Case "DrawLines" : GraphPaper1.GraphType = Val_Int(l)
                End Select
            Loop
            f.Close()
        End If
        LimitFormPosition(Form1)
        LimitFormPosition(Form2)
    End Sub


    Private Function ListToString(ByVal lst As List(Of String)) As String
        Dim str As String = ""
        For Each itm As String In lst
            If itm <> "" Then
                str += itm + ","
            End If
        Next
        If str.EndsWith(",") Then str = str.Substring(0, str.Length - 1)
        Return str
    End Function

    Private Function StringToList(ByVal str As String) As List(Of String)
        StringToList = New List(Of String)
        Dim sa As String() = str.Split(","c)
        For Each str In sa
            If str <> "" Then
                StringToList.Add(str)
            End If
        Next
    End Function

    ' ==================================================================================================
    '  SAVE LOAD -- OPTIONS FILE
    ' ==================================================================================================
    Friend Sub SaveOptionsFile()
        Dim sfd As SaveFileDialog = New SaveFileDialog()
        sfd.InitialDirectory = IO.Path.GetDirectoryName(OptionsFilePathAndName)
        sfd.DefaultExt = ".txt"
        sfd.AddExtension = True
        sfd.FileName = IO.Path.GetFileName(OptionsFilePathAndName)
        sfd.Filter = "Text file (*.txt)|*txt"
        IO.Directory.CreateDirectory(sfd.InitialDirectory)
        '
        SendKeys.Send("+{TAB}+{TAB}" + sfd.FileName + "{TAB}{TAB}") ' file selected in the list
        ''SendKeys.Send("{HOME}{END}^A") ' name visible and selected
        SendKeys.Send("{HOME}{END}")   ' name visible and unselected
        '
        If sfd.ShowDialog() = Windows.Forms.DialogResult.OK Then
            ' ---------------------------------------- save the old file before to change name and save the new file
            WriteOptionsFile()
            ' ---------------------------------------- save the new file
            OptionsFilePathAndName = sfd.FileName
            Form2.GroupBox1.Text = IO.Path.GetFileName(OptionsFilePathAndName)
            WriteOptionsFile()
            ReadDataFile(True)
            SetXaxisTimeUnit()
            Save_INI()
        End If
    End Sub

    Friend Sub SelectOptionsFile()
        Dim ofd As OpenFileDialog = New OpenFileDialog()
        ofd.InitialDirectory = IO.Path.GetDirectoryName(OptionsFilePathAndName)
        If Not IO.Directory.Exists(ofd.InitialDirectory) Then
            ofd.InitialDirectory = Application.StartupPath
        End If
        ofd.Multiselect = True
        ofd.Filter = "Options file (*.*)|*.*"
        ofd.DefaultExt = ".txt"
        IO.Directory.CreateDirectory(ofd.InitialDirectory)
        SendKeys.Send("+{TAB}+{TAB}" + EscapeSpecialCharacters(IO.Path.GetFileName(OptionsFilePathAndName)) + "{TAB}{TAB}")
        If ofd.ShowDialog() = Windows.Forms.DialogResult.OK Then
            OptionsFilePathAndName = ofd.FileName
            Form2.GroupBox1.Text = IO.Path.GetFileName(OptionsFilePathAndName)
            Form2.Options_FirstVisibleLine = 0
            ReadOptionsFile()
            ReadDataFile(True)
            SetXaxisTimeUnit()
            Save_INI()
        End If
    End Sub

End Module
