﻿
Module Module_Utils

    ' =============================================================================================
    '  InitPictureboxImage
    ' ============================================================================================= 
    Friend Sub InitPictureboxImage(ByVal pbox As PictureBox)
        With pbox
            If .ClientSize.Width < 1 Or .ClientSize.Height < 1 Then Return
            If .Image Is Nothing OrElse .Image.Width <> .ClientSize.Width _
                                 OrElse .Image.Height <> .ClientSize.Height Then
                .Image = New Bitmap(.ClientRectangle.Width, _
                                    .ClientRectangle.Height, _
                                    Imaging.PixelFormat.Format24bppRgb)
            End If
        End With
    End Sub

    ' =============================================================================================
    '  RemoveComments
    ' ============================================================================================= 
    Friend Function RemoveComments(ByVal s As String) As String
        Dim i As Int32 = s.IndexOf("'")
        If i > 0 Then s = s.Remove(i).TrimEnd
        If i = 0 Then s = ""
        Return s
    End Function

    ' =============================================================================================
    '  WINDOWS SCHEDULER PRECISION
    ' ============================================================================================= 
    Private Declare Function timeBeginPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32
    Private Declare Function timeEndPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32
    Friend Sub SchedulerMaxPrecision()
        timeBeginPeriod(1)
    End Sub
    Friend Sub SchedulerDefaultPrecision()
        timeEndPeriod(1)
    End Sub

    ' ==============================================================
    '  REPLACE MULTIPLE SPACES and TABS and TRIM  
    '  (3..5 uS for Strings with 32 chars)
    ' ============================================================== 
    Friend Function ReplaceMultipleSpacesAndTabAndTrim(ByVal s As String) As String
        If s = "" Then Return s
        s = s.Replace(vbTab, " ")
        While s.Contains("  ")
            s = s.Replace("  ", " ")
        End While
        Return s.Trim
    End Function

    ' ----------------------------------------------------------------------------------
    '  Extract from "text" start and length of the word pointed by the "selectionIndex"
    ' ----------------------------------------------------------------------------------
    Friend Sub FindWordPointers(ByVal text As String, _
                                ByVal selectionIndex As Int32, _
                                ByRef start As Int32, _
                                ByRef lenght As Int32)
        Const separators As String = " " + vbCr + vbLf + vbTab
        Dim i1, i2 As Int32
        For i1 = selectionIndex To 1 Step -1
            If i1 - 1 < text.Length AndAlso InStr(separators, text(i1 - 1)) > 0 Then
                Exit For
            End If
        Next
        For i2 = selectionIndex To text.Length - 1
            If InStr(separators, text(i2)) > 0 Then
                Exit For
            End If
        Next
        start = i1
        lenght = i2 - i1
    End Sub

    ' ----------------------------------------------------------------------------------
    '  Extract from "text" start and length of the word pointed by the "selectionIndex"
    '  This version find shortest words - not including characters like  " , / () ...
    ' ----------------------------------------------------------------------------------
    Friend Sub FindWordPointers_ShortWords(ByVal text As String, _
                                           ByVal selectionIndex As Int32, _
                                           ByRef start As Int32, _
                                           ByRef lenght As Int32)
        '
        Const separators As String = " "";:,/\|()'" + vbCr + vbLf + vbTab
        '
        Dim i1, i2 As Int32
        For i1 = selectionIndex To 1 Step -1
            If i1 - 1 < text.Length AndAlso InStr(separators, text(i1 - 1)) > 0 Then
                Exit For
            End If
        Next
        For i2 = selectionIndex To text.Length - 1
            If InStr(separators, text(i2)) > 0 Then
                Exit For
            End If
        Next
        start = i1
        lenght = i2 - i1
    End Sub

    ' ======================================================================================================
    '  ASYNC KEY STATE
    ' ======================================================================================================
    Private Declare Function GetAsyncKeyState Lib "user32" (ByVal vKey As Int32) As Int16
    Private Const PRESSED_NOW As Int32 = &H8000
    Private Const PRESSED_AFTER_PREVIOUS_CALL As Int32 = &H1

    Friend Function KeyPressedNow(ByVal k As Int32) As Boolean
        KeyPressedNow = (GetAsyncKeyState(k) And PRESSED_NOW) <> 0
    End Function

    Friend Function KeyPrevious(ByVal k As Int32) As Boolean
        KeyPrevious = (GetAsyncKeyState(k) And PRESSED_AFTER_PREVIOUS_CALL) <> 0
    End Function

    Private Declare Function VkKeyScan Lib "user32" Alias "VkKeyScanA" (ByVal chr As Int16) As Int16
    Friend Function KeyCodeToVirtualKey(ByVal k As Int32) As Int16
        If k > 65535 Then k = 0
        Return VkKeyScan(CShort(k))
    End Function

    Friend Sub WaitMouseOff()
        While KeyPressedNow(Keys.LButton) Or KeyPressedNow(Keys.MButton) Or KeyPressedNow(Keys.RButton)
            Threading.Thread.Sleep(1)
        End While
    End Sub

    Friend Sub WaitKeyOff(ByVal WaitKey As Int32)
        Do
            Threading.Thread.Sleep(1)
        Loop Until Not KeyPressedNow(WaitKey)
    End Sub


    ' =============================================================================================
    '  RELATIVE PATH / ABSOLUTE PATH
    ' =============================================================================================
    Friend Function RelativePath_FromAbsolute(ByVal AbsolutePath As String, _
                                              Optional ByVal BasePath As String = "") As String
        Try
            ' ----------------------------------------------------------- 
            If BasePath = "" Then BasePath = Application.StartupPath
            If Not BasePath.EndsWith("\") Then BasePath += "\"
            ' ----------------------------------------------------------- 
            Dim BaseUri As Uri = New Uri(BasePath)
            Dim FullUri As Uri = New Uri(AbsolutePath)
            ' ----------------------------------------------------------- 
            Dim RelativeUri As Uri = BaseUri.MakeRelativeUri(FullUri)
            ' ----------------------------------------------------------- 
            Dim RelativePath As String = RelativeUri.ToString().Replace("/", "\")
            RelativePath = Uri.UnescapeDataString(RelativePath)
            If Not RelativePath.StartsWith(".\") Then RelativePath = ".\" + RelativePath
            ' ----------------------------------------------------------- 
            Return RelativePath
        Catch
            Return Application.StartupPath
        End Try
    End Function

    Friend Function AbsolutePath_FromRelative(ByVal RelativePath As String, _
                                              Optional ByVal BasePath As String = "") As String
        ' ----------------------------------------------------------- 
        If BasePath = "" Then BasePath = Application.StartupPath
        If Not BasePath.EndsWith("\") Then BasePath += "\"
        ' -----------------------------------------------------------
        Return IO.Path.GetFullPath(IO.Path.Combine(BasePath, RelativePath))
    End Function


    ' =============================================================================================
    '  RoundValueTo_1_2_5
    ' ============================================================================================= 
    Friend Function RoundValueTo_1_2_5(ByVal number As Double) As Double
        If Double.IsNaN(number) Then Return 1
        Dim PowerOfTen As Double
        ' ------------------------------------------------------- find the nearest PowerOfTen
        PowerOfTen = Math.Round(Math.Log10(Math.Abs(number)))
        ' ------------------------------------------------------- Divide number by the PowerOfTen
        PowerOfTen = 10 ^ PowerOfTen
        Dim n As Double = Math.Abs(number) / PowerOfTen
        ' ------------------------------------------------------- test if result is near to 1, 2 or 5
        Dim n125 As Double
        If n <= 0.2 Then
            n125 = 0.1
        ElseIf n <= 0.5 Then
            n125 = 0.2
        ElseIf n <= 1 Then
            n125 = 0.5
        ElseIf n <= 2 Then
            n125 = 1
        ElseIf n <= 5 Then
            n125 = 2
        Else
            n125 = 5
        End If
        ' -------------------------------------------------------- create the final value
        Return n125 * PowerOfTen * Math.Sign(number)
    End Function

    ' =============================================================================================
    '  RoundToNdigits
    ' ============================================================================================= 
    Friend Function RoundToNdigits(ByVal n As Double, ByVal nDigits As Integer) As Double
        ' ----------------------------------------------------- can not calculate log(0)
        If Double.IsNaN(n) Then Return 0
        If Double.IsInfinity(n) Then Return 0
        If n = 0 Then Return 0
        ' ----------------------------------------------------- calculate shiftAmount
        Dim numDP As Int32 = CInt(nDigits - Math.Floor(Math.Log10(Math.Abs(n))) - 1)
        Dim shiftAmount As Double = Math.Pow(10, numDP)
        ' ----------------------------------------------------- shift up
        n = n * shiftAmount
        ' ----------------------------------------------------- nearest integer
        n = CInt(n)
        ' ----------------------------------------------------- shift down
        n = n / shiftAmount
        ' ----------------------------------------------------- return
        Return n
    End Function

    Friend Function RoundToNdigits2(ByVal n As Double, _
                                      ByVal nDigits As Integer, _
                                      ByVal nMaxDecimals As Integer, _
                                      ByVal LastSignificativeDigit As Double) As Double
        ' ----------------------------------------------------- can not calculate log(0)
        If Double.IsNaN(n) Then Return 0
        If Double.IsInfinity(n) Then Return 0
        If n = 0 Then Return 0
        ' ----------------------------------------------------- limit to last significative digit
        Dim magnitude As Integer = CInt(Math.Floor(Math.Log10(Math.Abs(n)))) - 1
        Dim magnitudeDiff As Integer = CInt(magnitude - Int(Math.Floor(Math.Log10(Math.Abs(LastSignificativeDigit)))))
        If nDigits > magnitudeDiff Then
            nDigits = magnitudeDiff
        End If
        ' ----------------------------------------------------- calculate shiftAmount
        Dim numDP As Integer = (nDigits - magnitude)
        If numDP > nMaxDecimals Then
            numDP = nMaxDecimals
        End If

        Dim shiftAmount As Double = Math.Pow(10, numDP)
        ' ----------------------------------------------------- shift up
        n = n * shiftAmount
        ' ----------------------------------------------------- nearest integer
        If (n > 0) Then
            n = Int(n + 0.5)
        Else
            n = Int(n - 0.5)
        End If
        ' ----------------------------------------------------- shift down
        n /= shiftAmount
        ' ----------------------------------------------------- return
        Return n
    End Function


    ' =============================================================================================
    '  EscapeSpecialCharacters for SendKeys
    ' ============================================================================================= 
    Friend Function EscapeSpecialCharacters(ByVal txt As String) As String
        txt = txt.Replace("+", "{+}")
        txt = txt.Replace("^", "{^}")
        txt = txt.Replace("%", "{%}")
        txt = txt.Replace("~", "{~}")
        txt = txt.Replace("(", "{(}")
        txt = txt.Replace(")", "{)}")
        Return txt
    End Function


End Module
